#include <cstdio>

using std::perror;

#include "reader.hh"
#include "defs.hh"

map<unsigned, Instr> Instrs;
map<unsigned, Sample> Samples;
Channel Channels[8];

static bool evcompare(const Event &a, const Event &b)
{
    if(a.time < b.time) return true;
    if(a.time > b.time) return false;
    // Always put 'F' and 'V' before 'O', if possible
    if(a.evtype == 'F' && b.evtype == 'O') return true;
    if(b.evtype == 'F' && a.evtype == 'O') return false;
    if(a.evtype == 'V' && b.evtype == 'O') return true;
    if(b.evtype == 'V' && a.evtype == 'O') return false;
    return false;
}

void ReadLog(FILE *fp)
{
    char Buf[512];
    while(fgets(Buf, sizeof Buf, fp))
    {
        char *s;
        if((s = strchr(Buf, '#'))) { for(*s=0; --s >= Buf; *s=0)if(*s!=' ')break; }
        switch(*Buf)
        {
            case 'I': // instrument
            {
                Instr tmp;
                unsigned insnum;
                sscanf(Buf, "I%u:%d %d %d %d %d %X",
                    &insnum,
                    &tmp.attack, &tmp.decay, &tmp.sustain, &tmp.release,
                    &tmp.sustainlevel, &tmp.sampleid);
                Instrs[insnum] = tmp;
                break;
            }
            case 'S': // sample
            {
                unsigned smpid;
                unsigned loopbeg, loopend;
                int state;
                sscanf(Buf, "S%X:%u-%u %d", &smpid, &loopbeg, &loopend, &state);

                Sample &smp = Samples[smpid];
                smp.looping = state >= 2;
                smp.loopbeg = loopbeg;
                smp.loopend = loopend;
                break;
            }
            case 'D': // sample file
            {
                unsigned smpid;
                sscanf(Buf, "D%X:%s", &smpid, Buf);
                FILE *fp = fopen(Buf, "rb");
                if(!fp)
                {
                	perror(Buf);
                	break;
                }
                fseek(fp, 0, SEEK_END); unsigned long fsize = ftell(fp); rewind(fp);
                char *data = new char[fsize];
                fread(data, 1, fsize, fp);
                fclose(fp);
                Samples[smpid].data.assign(data, fsize);
                delete[] data;
                break;
            }
            case 'V': // volumes
            {
                Event tmp('V');
                unsigned ch;
                sscanf(Buf, "V%u@%lf:%d %d", &ch, &tmp.time, &tmp.l, &tmp.r);
                /*
                printf("From '%s' got ch=%u, time=%g, l=%d, r=%d\n",
                    Buf, ch, tmp.time, tmp.l, tmp.r);
                */
                if(tmp.l < 0)tmp.l = -tmp.l;
                if(tmp.r < 0)tmp.r = -tmp.r;
                Channels[ch].events.push_back(tmp);
                break;
            }
            case 'F': // freq
            {
                Event tmp('F');
                unsigned ch;
                sscanf(Buf, "F%u@%lf:%lu", &ch, &tmp.time, &tmp.freq);
                Channels[ch].events.push_back(tmp);
                break;
            }
            case 'K': // key off
            {
                Event tmp('K');
                unsigned ch;
                sscanf(Buf, "K%u@%lf", &ch, &tmp.time);
                Channels[ch].events.push_back(tmp);
                break;
            }
            case 'O': // key on
            {
                Event tmp('O');
                unsigned ch;
                sscanf(Buf, "O%u@%lf", &ch, &tmp.time);
                Channels[ch].lastkeyon = Channels[ch].events.size();
                Channels[ch].events.push_back(tmp);
                break;
            }
            case 'W': // instrument was
            {
                unsigned ch;
                unsigned ins;
                sscanf(Buf, "W%u:%u", &ch, &ins);
                unsigned lastkeyon = Channels[ch].lastkeyon;
                Event &tmp = Channels[ch].events[lastkeyon];
                tmp.ins = ins;
                break;
            }
        }
    }
    for(unsigned ch=0; ch<8; ++ch)
    {
        sort(Channels[ch].events.begin(),
             Channels[ch].events.end(),
             evcompare);
    }
}
